// ==============================================================
//	This file is part of Glest Shared Library (www.glest.org)
//
//	Copyright (C) 2001-2007 Martio Figueroa
//
//	You can redistribute this code and/or modify it under
//	the terms of the GNU General Public License as published
//	by the Free Software Foundation; either version 2 of the
//	License, or (at your option) any later version
// ==============================================================

#include "pch.h"
#include "properties.h"

#include <fstream>
#include <stdexcept>
#include <cstring>
#include <sstream>
#include <locale>

#include "conversion.h"

#include "leak_dumper.h"
#include "FSFactory.hpp"
using std::exception;

namespace Shared { namespace Util {

using namespace PhysFS;

// =====================================================
// class Properties
// =====================================================
	
Properties::Properties() {}

void Properties::load(const string &path, bool trim) {
	std::locale loc;
	char lineBuffer[maxLine];
	string line, key, value;
	int pos;

	this->path = path;

	istream *fileStream = g_fileFactory.getIStream(path.c_str());
	//fileStream.exceptions(ios::failbit | ios::badbit);

	if (fileStream->fail()) {
		throw runtime_error("Can't find or open propertyMap file: " + path);
	}

	propertyMap.clear();

	while (!fileStream->eof()) {
		fileStream->getline(lineBuffer, maxLine);

		lineBuffer[maxLine - 1] = '\0';

		if (lineBuffer[0] == ';' || lineBuffer[0] == '#') {
			continue;
		}

		// gracefully handle win32 \r\n line endings
		size_t len = strlen(lineBuffer);
		if (len > 0 && lineBuffer[len - 1] == '\r') {
			lineBuffer[len - 1] = 0;
		}

		line = lineBuffer;
		pos = line.find('=');

		if (pos == string::npos) {
			continue;
		}

		key = line.substr(0, pos);
		value = line.substr(pos + 1);

		if (trim) {
			key = Conversion::trimStr(key);
			value = Conversion::trimStr(value);
		}
		for (int i=0; i < key.size(); ++i) {
			key[i] = tolower(key[i]);
		}
		propertyMap.insert(PropertyPair(key, value));
		propertyVector.push_back(PropertyPair(key, value));
	}

	delete fileStream;
}

void Properties::save(const string &path) {
	ostream *fileStream = FSFactory::getInstance()->getOStream(path.c_str());

	*fileStream << "; === propertyMap File === \n";
	*fileStream << '\n';

	for (PropertyMap::iterator pi = propertyMap.begin(); pi != propertyMap.end(); ++pi) {
		*fileStream << pi->first << '=' << pi->second << '\n';
	}

	delete fileStream;
}

void Properties::clear(){
	propertyMap.clear();
	propertyVector.clear();
}

const string *Properties::_getString(const string &key, bool required) const {
	PropertyMap::const_iterator it;
	string lkey = key;
	for (int i=0; i < lkey.size(); ++i) {
		if (isupper(lkey[i])) {
			lkey[i] = tolower(lkey[i]);
		}
	}
	it = propertyMap.find(lkey);
	if (it == propertyMap.end()) {
		if (required) {
			throw runtime_error("Value not found in propertyMap: " + key + ", loaded from: " + path);
		}
		return 0;
	} else {
		return &it->second;
	}
}

template<typename T>
static inline void checkRange(const T &val, const T *pMin, const T *pMax) {
	if((pMin && val < *pMin) || (pMax && val > *pMax)) {
		stringstream str;
		str << "Value out of range: " << val << "(min = ";
		if(pMin) {
			str << *pMin;
		} else {
			str << "none";
		}
		str << ", max = ";
		if(pMax) {
			str << *pMax << ")";
		} else {
			str << "none)";
		}
		throw range_error(str.str());
	}
}

bool Properties::_getBool(const string &key, const bool *pDef) const {
	//try {
		const string *pstrVal = _getString(key, bool(!pDef));
		if (!pstrVal) {
			assert(pDef);
			return *pDef;
		}

		return Conversion::strToBool(*pstrVal);
	//} catch (exception &e) {
	//	throw runtime_error("Error accessing value: " + key + " in: " + path + "\n" + e.what());
	//}
}

int Properties::_getInt(const string &key, const int *pDef, const int *pMin, const int *pMax) const {
	//try {
		const string *pstrVal = _getString(key, !pDef);
		if(!pstrVal) {
			assert(pDef);
			return *pDef;
		}

		int val =  Conversion::strToInt(*pstrVal);
		checkRange<int>(val, pMin, pMax);
		return val;
	//} catch (exception &e) {
	//	throw runtime_error("Error accessing value: " + key + " in: " + path + "\n" + e.what());
	//}
}

float Properties::_getFloat(const string &key, const float *pDef, const float *pMin, const float *pMax) const {
	//try {
		const string *pstrVal = _getString(key, !pDef);
		if(!pstrVal) {
			assert(pDef);
			return *pDef;
		}
		float val =  Conversion::strToFloat(*pstrVal);
		checkRange<float>(val, pMin, pMax);
		return val;
	//} catch (exception &e) {
	//	throw runtime_error("Error accessing value: " + key + " in: " + path + "\n" + e.what());
	//}
}

const string &Properties::_getString(const string &key, const string *pDef) const {
	//try {
		const string *pstrVal = _getString(key, !pDef);
		if(!pstrVal) {
			assert(pDef);
			return *pDef;
		}
		return *pstrVal;
	//} catch (exception &e) {
	//	throw runtime_error("Error accessing value: " + key + " in: " + path + "\n" + e.what());
	//}
}

string Properties::toString() const {
	stringstream str;

	for (PropertyMap::const_iterator pi = propertyMap.begin(); pi != propertyMap.end(); ++pi)
		str << pi->first << "=" << pi->second << endl;

	return str.str();
}

}}//end namepsace
